/*
** File: packages/curl/curlconfig.P
** Author: Aneesh Ali
** Contact:   xsb-contact@cs.sunysb.edu
** 
** Copyright (C) The Research Foundation of SUNY, 2010
** 
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
*/


:- compiler_options([xpp_on]).
#include "extensions_xsb.h"


:- import slash/1, str_cat/3 from machine.
:- import search_module/6 from consult.
:- import path_sysop/3, path_sysop/2, fmt_write_string/3 from file_io.
:- import xsb_configuration/2 from xsb_configuration.
:- import concat_atom/2 from string.

'_$_curl_make_curlH'(CC, CC_opts) :-
	slash(Slash),
	xsb_configuration(configuration, XSBconfiguration),
	(('_$_curl_running_under'(unix) ; '_$_curl_running_under'(darwin)) ->
	    xsb_configuration(loader_flags, LoaderFlagsBase), 
	    concat_atom([LoaderFlagsBase, ' -lcurl'], LoaderFlags),
	    LD_flags = LoaderFlags
	; %% Cygwin. Under Windows this option is ignored
	    LD_flags = ' -lwsock32 '
	),

	package_configuration(dir(curl), CurlDir),
	fmt_write_string(LD_directives_file,
			 '%s%scc%scurl2pl.H', f(CurlDir, Slash, Slash)),
	telling(OldTell),
	tell(LD_directives_file),
	writeln('%% This file is generated on-the-fly by curlconfig.P'),
	writeln('%% Do not edit--any changes will be written over.'),
	write('%% config:    '), writeln(XSBconfiguration),
	write('%% CC:        '), writeln(CC),
	write('%% CC Opts: '), writeln(CC_opts), 
	writeln(':- export pl_load_page/1, curl_allocate_error_term/2, curl_finalize_warn/1, pl_encode_url/2.'),
	write(':- ldoption('''),
	write(LD_flags),
	writeln(''').'),
	told,
	tell(OldTell).


:-
    package_configuration(dir(curl), CurlDir),  
    slash(Slash),
    Basename = curl2pl,
    xsb_configuration(config_libdir, ConfigLibDir),
    xsb_configuration(compiler, CC),
    xsb_configuration(config_dir, ConfigDir),
    xsb_configuration(compiler_flags, CCflags),
    xsb_configuration(config_bindir, ConfigBinDir),

    ('_$_curl_running_under'(windows64) -> PkgDlls = bin64
    ; PkgDlls = bin
    ),
    fmt_write_string(WindCurlDlls,
		     '%s%scc%s%s%s*.dll', f(CurlDir,Slash,Slash,PkgDlls,Slash)),
    fmt_write_string(WindCurlLibs,
		     '%s%scc%s%s%s*.lib', f(CurlDir,Slash,Slash,PkgDlls,Slash)),

    fmt_write_string(Module, '%s%scc%s%s', f(CurlDir,Slash,Slash,Basename)),
    fmt_write_string(LibObjFile,
		     '%s%s%s%s',
		     args(ConfigLibDir,Slash,Basename,XSB_OBJ_EXTENSION_ATOM)),
    fmt_write_string(HFile, '%s%scc%sload_page.c', 
				 f(CurlDir, Slash, Slash)),
    fmt_write_string(CC_opts, ' %s -I %s %s', f(CCflags, ConfigDir, HFile)),
    str_cat(Module, XSB_OBJ_EXTENSION_ATOM, Objfile),
    (('_$_curl_running_under'(windows) ; '_$_curl_running_under'(cygwin))
    ->
	fmt_write_string(ConfigSharedMod,
			 '%s%s%s.dll', f(ConfigBinDir, Slash, Basename))
    ; '_$_curl_running_under'(darwin)
    ->
	fmt_write_string(ConfigSharedMod,
			 '%s%s%s.dylib', f(ConfigLibDir, Slash, Basename))
    ;
	fmt_write_string(ConfigSharedMod,
			 '%s%s%s.so',  f(ConfigLibDir, Slash, Basename))
    ),
    (   %% avoid recompilation
	search_module(Basename,Dir,_Mod,_Ext,BaseDir,ModuleO),
	str_cat(Module, '.c', ModuleSourceFile),
	str_cat(ConfigBinDir, 'libcurl.dll', LibcurlDLL),
	str_cat(ConfigBinDir, 'curl2pl.dll', Curl2plDLL),
	path_sysop(newerthan,ModuleO,ModuleSourceFile),
	path_sysop(newerthan,ConfigSharedMod,ModuleSourceFile),
	(('_$_curl_running_under'(windows)
	 ; '_$_curl_running_under'(windows))
	->
	    path_sysop(exists,LibcurlDLL)
	; true
	)
    %% module compiled and is current, so just load
    -> [Basename]
    
    %% If Module isn't compiled or is old --- recompile
    ;
	'_$_curl_make_curlH'(CC, CC_opts),
	('_$_curl_running_under'(windows) ->
	    shell(['copy /B /Y ', WindCurlDlls, ' ', ConfigBinDir]),
	    shell(['copy /B /Y ', WindCurlLibs, ' ', ConfigBinDir]),
	    [Basename],
	    (path_sysop(exists,Objfile) -> '_$_curl_force_rename'(Objfile, LibObjFile)
	    ; true
	    )

	; '_$_curl_running_under'(cygwin) ->
	    shell(['copy /B /Y ', WindCurlDlls, ' ', ConfigBinDir]),
	    shell(['copy /B /Y ', WindCurlLibs, ' ', ConfigBinDir]),

	    str_cat(Module, '.dll', SharedLib),
	    consult(Module, [cc(CC), cc_opts(CC_opts)]),
	    '_$_curl_force_rename'(SharedLib, ConfigSharedMod),
	    '_$_curl_force_rename'(Objfile, LibObjFile),
	    str_cat(Module, '.o', ObjfileO),
	    fmt_write_string(LibObjFileO,
			     '%s%s%s.o', f(ConfigBinDir, Slash, Basename)),
	    '_$_curl_force_rename'(ObjfileO, LibObjFileO),
	    str_cat(Module, '.a', ObjfileA),
	    fmt_write_string(LibObjFileA,
			     '%s%s%s.a', f(ConfigBinDir, Slash, Basename)),
	    '_$_curl_force_rename'(ObjfileA, LibObjFileA),
	    str_cat(Module, '.def', ObjfileDef),
	    fmt_write_string(LibObjFileDef,
			     '%s%s%s.def', f(ConfigBinDir, Slash, Basename)),
	    '_$_curl_force_rename'(ObjfileDef, LibObjFileDef)

	; '_$_curl_running_under'(darwin) ->  %% MacOS
	    consult(Module, [cc(CC), cc_opts(CC_opts)]),
	    '_$_curl_force_rename'(Objfile, LibObjFile),
	    str_cat(Module, '.dylib', SharedLib),
	    '_$_curl_force_rename'(SharedLib,ConfigSharedMod)
	; %% Other Unices
	    consult(Module, [cc(CC), cc_opts(CC_opts)]),
	    '_$_curl_force_rename'(Objfile, LibObjFile),
	    str_cat(Module, '.so', SharedLib),
	    '_$_curl_force_rename'(SharedLib,ConfigSharedMod)
	)
    ).

%% Windows semantics for renaming is such that the op fails, 
%% if the target exists
'_$_curl_force_rename'(Orig,Target) :-
	(path_sysop(unlink,Target), !; true),
	path_sysop(rename,Orig,Target).


'_$_curl_running_under'(cygwin) :-
	xsb_configuration(architecture, A),
	str_sub(cygwin, A),
	!.
'_$_curl_running_under'(windows) :-
	xsb_configuration(architecture, A),
	str_sub(windows, A),
	!.
'_$_curl_running_under'(windows64) :-
	xsb_configuration(architecture, A),
	str_sub(windows, A),
	str_sub(x64, A),
	!.
'_$_curl_running_under'(darwin) :-
	xsb_configuration(architecture, A),
	str_sub(darwin, A),
	!.
'_$_curl_running_under'(unix) :-
	\+ '_$_curl_running_under'(cygwin),
	\+ '_$_curl_running_under'(darwin),
        \+ '_$_curl_running_under'(windows).
